/*
-------------------------------------------------------------

                 Probabilistic Decision Diagrams

                        Implementation
-------------------------------------------------------------
-------------------------------------------------------------
        Module              :           %M%
        Date                :           %G%
        Version             :           %I%
        Author              :           Dorel Marius Bozga
-------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pdd.h"

#define pdd_message(m)\
  fprintf(stderr,"\tpdd : %s\n",m)

#define pdd_abort()\
  exit(0)



#define pdd_set_mark(p)\
  (p)->flags |= 1

#define pdd_reset_mark(p)\
  (p)->flags &= (~1)

#define pdd_test_mark(p)\
  ((p)->flags & 1)

/*
-------------------------------------------------------------
*/

void
pdd_create_manager(pdd_manager** m)
{
  int i;
  (*m) = (pdd_manager*)malloc(sizeof(pdd_manager));
  (*m)->size = 16381;
  (*m)->entries = (pdd**) malloc(sizeof(pdd*) * (*m)->size);
  for(i = 0; i < (*m)->size; i++)
    (*m)->entries[i] = NULL;
}

void
pdd_delete_manager(pdd_manager** m)
{
  int i;
  for(i = 0; i < (*m)->size; i++){
    pdd* crt, *next;
    for(crt = (*m)->entries[i]; crt!=NULL; ){
      next = crt->next;
      free(crt);
      crt = next;
    }
  }
  free((*m)->entries);
  free((*m));
  *m = NULL;
}

/*
-------------------------------------------------------------
*/

pdd* 
pdd_mknode(pdd_manager* m, float proba, int index, pdd* left, pdd* right){
  int hash = ((((long) index) << 7) + 
	      (((long) left) << 3 ) + 
	      (((long) right) << 5 )) % m->size;
  pdd* node = NULL;
  for(node = m->entries[hash]; node != NULL; node = node->next)
    if (node->proba == proba &&
	node->index == index && 
	node->left == left && 
	node->right == right)
      break;
  if (node == NULL){
    node = (pdd*) malloc(sizeof(pdd));
    node->proba = proba;
    node->index = index;
    node->left = left;
    node->right = right;
    node->flags = 0;
    node->next = m->entries[hash];
    m->entries[hash] = node;
  }
  return node;
}

/*
-------------------------------------------------------------
*/

void 
pdd_mark(pdd* p){
  if (! pdd_test_mark(p)){
    pdd_set_mark(p);
    if (p->left)
      pdd_mark(p->left);
    if (p->right)
      pdd_mark(p->right);
  }
}


/*
-------------------------------------------------------------
*/

void
pdd_size_aux(pdd* p, int* counter){
  if (pdd_test_mark(p)){
    if (p->left)
      pdd_size_aux(p->left,counter);
    if (p->right)
      pdd_size_aux(p->right,counter);
    pdd_reset_mark(p);
    (*counter)++;
  }
}

int
pdd_size(pdd* p){
  int size = 0;
  pdd_mark(p);
  pdd_size_aux(p, &size);
  return size;
}

/*
-------------------------------------------------------------
*/

void 
pdd_print_aux(pdd* p, FILE* file){
  if (pdd_test_mark(p)){
    if (p->left)
      pdd_print_aux(p->left, file);
    if (p->right)
      pdd_print_aux(p->right, file);
    pdd_reset_mark(p);
    fprintf(file,"%#x : <proba> %.3f <var> x%-3d <left> %#x <right> %#x\n",
	    p, p->proba, p->index, p->left, p->right);
  }
}

void 
pdd_print(pdd* p, FILE* file){
  pdd_mark(p);
  pdd_print_aux(p, file);
}

/*
-------------------------------------------------------------
*/

void 
pdd_print_tree_aux(pdd* p, FILE* file, int depth, char* path){
  int i;
  if (p->left != NULL){
    path[depth] = 0;
    pdd_print_tree_aux(p->left, file, depth + 1, path);
  }

  for(i = 0; i < depth; i++)
    fprintf(file,"%c        ", i && path[i] != path[i-1] ? '|' : ' ');
  fprintf(file, "+x%-3d%4.2f|\n", p->index, p->proba);

  if (p->right != NULL){
    path[depth] = 1;
    pdd_print_tree_aux(p->right, file, depth + 1, path);
  }
}

void 
pdd_print_tree(pdd* p, FILE* file){
  char path[128] = "";
  pdd_print_tree_aux(p, file, 0, path);
}


/*
-------------------------------------------------------------
*/

pdd* 
pdd_create_aux(pdd_manager* m, float* proba, int size, int index, float value){
  pdd* node;
  if (size == 1)
    node = pdd_mknode(m, value, index, 
		      NULL, 
		      NULL);
  else {
    float s1 = 0.0, s2 = 0.0, *p = proba;
    int i;
    for(i = 0; i < size / 2; i++)
      s1 += *p++;
    for(i = 0; i < size / 2; i++)
      s2 += *p++;
    node = pdd_mknode(m, value, index, 
		      pdd_create_aux(m, proba, size / 2, index + 1, s1 / (s1 + s2)),
		      pdd_create_aux(m, proba + size / 2, size / 2, index + 1, s2 / (s1 + s2)));
  }
  return node;
}

pdd*
pdd_create(pdd_manager* m, float* proba, int size){
  return pdd_create_aux(m, proba, size, 0, 1.0);
}


/*
-------------------------------------------------------------
*/

int read_proba(FILE* file, float* prob){
  int size = 0;
  if (fscanf(file, "%d", &size) == 1){
    int i;
    for(i = 0; i < size; i++)
      if (fscanf(file, "%f", &prob[i]) != 1){
	size = 0;
	break;
      }
  }
  return size;
}

/*
-------------------------------------------------------------
*/

void 
main()
{
  float proba[64];
  int size = 0;

  pdd_manager* m;
  pdd_create_manager(&m);
  
  size = read_proba(stdin, proba);
  if (size){
    pdd* p = pdd_create(m, proba, size);
    pdd_print_tree(p, stdout);
  }

  pdd_delete_manager(&m);
}
